/* 
 * virusscanning VFS module for samba.  Log infected files via syslog
 * facility and block access using SAVI.
 *
 * Copyright (C) Rainer Link, 2001-2002 
 *               OpenAntiVirus.org <rainer@openantivirus.org>
 *
 * 
 * based on the audit VFS module by
 * Copyright (C) Tim Potter, 1999-2000
 *
 * based on Sophie by
 * Copyright (C) Vanja Hrustic, 2001
 *
 * Credits to
 * - Dave Collier-Brown for his VFS tutorial (http://www.geocities.com/orville_torpid/papers/vfs_tutorial.html)
 * - REYNAUD Jean-Samuel for helping me to solve some general Samba VFS issues at the first place
 * - Simon Harrison for his solution without Samba VFS (http://www.smh.uklinux.net/linux/sophos.html)
 * - the whole Samba Team :)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <includes.h>
#include <vfs.h>

#include "vscan-sophos.h"


/* module version */
static const char module_id[]="vscan-sophos 0.2.5a";

extern bool savi_ok;

/* this is used in the hacked vscan_open() function 
 * Added by Kevin Spicer July 2002
 * Code pinched from samba honeypot code by Michael Gerdts
 * All errors probably mine! */
static char *mystrchr(char *path, int c) {
	char *p;
	for ( p = path; p != NULL && *p != '\0' ; p++ ) {
		if ( *p == c ) {
			return(p);
		}
	} return(NULL);
}

#if SAMBA_VERSION_MAJOR!=3
 /* Samba 2.2.x */
 /* Function prototypes */

 #if SAMBA_VERSION_MINOR==4
  /* Samba 2.2.4 */
  int vscan_connect(struct connection_struct *conn, const char *svc, const char *user);
  void vscan_disconnect(struct connection_struct *conn);
  int vscan_open(struct connection_struct *conn, const char *fname, int flags, mode_t mode);
  int vscan_close(struct files_struct *fsp, int fd);
 #else
  /* Samba 2.2.3-2.2.0 */
  int vscan_connect(struct connection_struct *conn, char *svc, char *user);
  void vscan_disconnect(struct connection_struct *conn);
  int vscan_open(struct connection_struct *conn, char *fname, int flags, mode_t mode);
  int vscan_close(struct files_struct *fsp, int fd);
 #endif
#endif
/* VFS operations */

#if SAMBA_VERSION_MAJOR!=3
 /* Samba 2.2.x */
 extern struct vfs_ops default_vfs_ops;   /* For passthrough operation */
#else
 /* Samba 3.0 alphaX */
 static struct vfs_ops default_vfs_ops;   /* For passthrough operation */
#endif

#if SAMBA_VERSION_MAJOR!=3 
 /* Samba 2.2.x */
 #if SAMBA_VERSION_MINOR==4
 /* Samba 2.2.4 */
 struct vfs_ops vscan_ops = {
    
	/* Disk operations */

	vscan_connect,			/* connect */
	vscan_disconnect,		/* disconnect */
	NULL,				/* disk free */

	/* Directory operations */

	NULL,				/* opendir */
	NULL,				/* readdir */
	NULL,				/* mkdir */
	NULL,				/* rmdir */
	NULL,				/* closedir */

	/* File operations */

	vscan_open,			/* open */
	vscan_close,			/* close */
	NULL,				/* read  */
	NULL,				/* write */
	NULL,				/* lseek */
	NULL,				/* rename */
	NULL,				/* fsync */
	NULL,				/* stat  */
	NULL,				/* fstat */
	NULL,				/* lstat */
	NULL,				/* unlink */
	NULL,				/* chmod */
	NULL,				/* fchmod */
	NULL,				/* chown */
	NULL,				/* fchown */
	NULL,				/* chdir */
	NULL,				/* getwd */
	NULL,				/* utime */
	NULL,				/* ftruncate */
	NULL,				/* lock */
	NULL,				/* symlink */
	NULL,				/* readlink */
	NULL,				/* link */
	NULL,				/* mknod */
	NULL,				/* realpath */
	NULL,				/* fget_nt_acl */
	NULL,				/* get_nt_acl */
	NULL,				/* fset_nt_acl */
	NULL,				/* set_nt_acl */

	NULL,				/* chmod_acl */
	NULL,				/* fchmod_acl */

	NULL,				/* sys_acl_get_entry */
	NULL,				/* sys_acl_get_tag_type */
	NULL,				/* sys_acl_get_permset */
	NULL,				/* sys_acl_get_qualifier */
	NULL,				/* sys_acl_get_file */
	NULL,				/* sys_acl_get_fd */
	NULL,				/* sys_acl_clear_perms */
	NULL,				/* sys_acl_add_perm */
	NULL,				/* sys_acl_to_text */
	NULL,				/* sys_acl_init */
	NULL,				/* sys_acl_create_entry */
	NULL,				/* sys_acl_set_tag_type */
	NULL,				/* sys_acl_set_qualifier */
	NULL,				/* sys_acl_set_permset */
	NULL,				/* sys_acl_valid */
	NULL,				/* sys_acl_set_file */
	NULL,				/* sys_acl_set_fd */
	NULL,				/* sys_acl_delete_def_file */
	NULL,				/* sys_acl_get_perm */
	NULL,				/* sys_acl_free_text */
	NULL,				/* sys_acl_free_acl */
	NULL				/* sys_acl_free_qualifier */
 };
 #else
 /* Samba 2.2.3-2.2.0 */
 struct vfs_ops vscan_ops = {
    
	/* Disk operations */

	vscan_connect,		  /* connect */
	vscan_disconnect,	  /* disconnect */
	NULL,                     /* disk free */

	/* Directory operations */

	NULL,			  /* opendir */
	NULL,                     /* readdir */
	NULL,			  /* mkdir */
	NULL,			  /* rmdir */	
	NULL,                     /* closedir */

	/* File operations */

	vscan_open,		  /* open  */
	vscan_close,		  /* close */
	NULL,                     /* read  */
	NULL,                     /* write */
	NULL,                     /* lseek */
	NULL,			  /* rname */
	NULL,                     /* fsync */
	NULL,                     /* stat  */
	NULL,                     /* fstat */
	NULL,                     /* lstat */
	NULL,			  /* unlink */
	NULL,			  /* chmod */
	NULL,                     /* chown */
	NULL,                     /* chdir */
	NULL,                     /* getwd */
	NULL,                     /* utime */
	NULL,                     /* ftruncate */
	NULL,                     /* lock */
	NULL,                     /* fget_nt_acl */
	NULL,                     /* get_nt_acl */
	NULL,                     /* fset_nt_acl */
	NULL                      /* set_nt_acl */
 };
 #endif
#endif 

/* Implementation of vfs_ops.  */

#if SAMBA_VERSION_MAJOR==3
 /* Samba 3.0 alphaX */
 static int vscan_connect(struct connection_struct *conn, const char *svc, const char *user)
#else
 /* Samba 2.2.x */
 #if SAMBA_VERSION_MINOR==4
  /* Samba 2.2.4 */
  int vscan_connect(struct connection_struct *conn, const char *svc, const char *user)
 #else
  /* Samba 2.2.3-2.2.0 */
  int vscan_connect(struct connection_struct *conn, char *svc, char *user)
 #endif
#endif
{
	vscan_syslog("connect to service %s by user %s\n", 
	       svc, user);

	vscan_sophos_init();

	#if SAMBA_VERSION_MAJOR!=3
	return default_vfs_ops.connect(conn, svc, user);
	#else
	return 0;	/* Success */
	#endif
}

#if SAMBA_VERSION_MAJOR==3
 /* Samba 3.0 alphaX */
 static void vscan_disconnect(struct connection_struct *conn)
#else
 /* Samba 2.2.x */
 void vscan_disconnect(struct connection_struct *conn)
#endif
{
        if ( savi_ok )
		vscan_sophos_end();

	/* Message changed by Kevin Spicer 9 July 2002 to reflect
	 * that session might not actually have disconnected */
	vscan_syslog("Disconnection request processed\n");

	#if SAMBA_VERSION_MAJOR!=3
	default_vfs_ops.disconnect(conn);
	#endif
}


#if SAMBA_VERSION_MAJOR==3
 /* Samba 3.0 alphaX */
 static int vscan_open(struct connection_struct *conn, const char *fname, int flags, mode_t mode)
#else
 /* Samba 2.2.x */
 #if SAMBA_VERSION_MINOR==4
  /* Samba 2.2.4 */
  int vscan_open(struct connection_struct *conn, const char *fname, int flags, mode_t mode)
 #else
  /* Samba 2.2.3 - 2.2.0 */
  int vscan_open(struct connection_struct *conn, char *fname, int flags, mode_t mode)
 #endif
#endif
{

//	#if SCAN_ON_OPEN==1
//	int retval;
//	SMB_STRUCT_STAT stat_buf;
//	char client_ip[16];
//
//	if ( !savi_ok ) {
//		/* SAVI initialisation failed - no virus scanning possible */
//		#if DENY_ACCESS_ON_ERROR==1
//              vscan_syslog("ERROR: SAVI initialisation failed, no virus scanning possible - access to file %s denied", fname);
//                errno = EACCES;
//                return -1;
//                #endif
//	} else {
//	        if ( (default_vfs_ops.stat(conn, fname, &stat_buf)) != 0 )    /* an error occured */ 
//			return default_vfs_ops.open(conn, fname, flags, mode);
//		else if ( S_ISDIR(stat_buf.st_mode) ) 	/* is it a directory? */
//			 return default_vfs_ops.open(conn, fname, flags, mode);
//		else if ( ( stat_buf.st_size > VSCAN_SOPHOS_MAX_SIZE ) && ( VSCAN_SOPHOS_MAX_SIZE > 0 ) ) /* file is too large */
//			vscan_syslog("File %s too large! Not scanned!", fname);
//		else if ( stat_buf.st_size == 0 ) /* do not scan empty files */
//			return default_vfs_ops.open(conn, fname, flags, mode);
//		else { 
//			fstrcpy(client_ip,conn->client_address);
//			retval = vscan_sophos_scanfile(fname, client_ip);
//			if ( retval == -1 ) {
//				/* an error occured during virus scan */
//				#if DENY_ACCESS_ON_ERROR==1
//		                vscan_syslog("ERROR: an error occured during virus scan - access to file %s denied", fname);
  //              		errno = EACCES;
//		                return -1;
//		                #endif
//			} else if ( retval == 1 ) {
//				errno = EACCES; 
//				return -1;
//			}
//		}
//	}
//	#endif

	/* A hack by Kevin Spicer July 2002
	 * to make the file really in a quaratine area 
	 * as these shares should not really be writable 
	 * Based on code by Mike Gerdts (Michael.Gerdts@usa.alcatel.com)
	 * All errors probably mine!*/
	
	
char qfname[PATH_MAX+1];
char qfdir[PATH_MAX+1];
char copycommand[PATH_MAX*3];
struct stat statbuf;
char *p, *file;

#if VSCAN_SOPHOS_VERBOSE_FILE_LOGGING==1
vscan_syslog("quarantine_open (%s, 0%o, 0%o)\n", fname, flags, mode);
#endif

if (flags & (O_WRONLY | O_RDWR | O_CREAT | O_TRUNC | O_APPEND)) {
	if (snprintf(qfname, PATH_MAX+1, "/quarantine/%s/%s/%s/%s",
		conn->client_address, conn->user,
		conn->origpath, fname)
			> PATH_MAX) {return(ENAMETOOLONG);
			}
	
	#if VSCAN_SOPHOS_VERBOSE_FILE_LOGGING==1
	vscan_syslog("quarantine_open: quarantine file is \"%s\"", qfname);
	#endif
	strncpy(qfdir, qfname, PATH_MAX+1);
	p = qfdir;
	while ((p=mystrchr(p, '/')) !=NULL) {
		*p='\0';
		if (stat(qfdir, &statbuf) != 0) {
			#if VSCAN_SOPHOS_VERBOSE_FILE_LOGGING==1
			vscan_syslog("quarantine_open: mkdir \"%s\"", qfdir);
			#endif
			mkdir(qfdir, 01777);
		}
		*p ='/';
		p++;
	}
	/* Copy the file for editing */
	if ( (flags & (O_WRONLY | O_RDWR)) && ! (flags & O_CREAT) && ! (flags & O_TRUNC) ) {
		/*This really is a dirty hack,  There must be a better way! */
		snprintf(copycommand, PATH_MAX*3, "/bin/cp /%s/%s /quarantine/%s/%s/%s/%s",
			conn->origpath,fname,
			conn->client_address, conn->user,
			conn->origpath, fname);
		#if VSCAN_SOPHOS_VERBOSE_FILE_LOGGING==1
		vscan_syslog("quarantine_open: copy command \"%s\"", copycommand);
		#endif
		
		system(copycommand);


	}
	file = qfname;
}
else {
	file = fname;
}
#if VSCAN_SOPHOS_VERBOSE_FILE_LOGGING==1
vscan_syslog("quarantine_open: open (\"%s\", 0%o, 0%o)", file, flags, mode);
#endif


	return default_vfs_ops.open(conn, file, flags, mode);

}


#if SAMBA_VERSION_MAJOR==3
 /* Samba 3.0 alphaX */
 static int vscan_close(struct files_struct *fsp, int fd)
#else
 /* Samba 2.2.x */
 int vscan_close(struct files_struct *fsp, int fd)
#endif
{
	#if SCAN_ON_CLOSE==1
        pstring filepath;
        int retval, rv, retval2;
	char client_ip[16];

        /* First close the file */
        retval = default_vfs_ops.close(fsp, fd);

	/* get the file name 
	 * Modified to pick up the path where the file actually is
	 * not where the client thinks it is*/
        //fstrcpy(filepath, fsp->conn->connectpath);
        //fstrcat(filepath, "/"); 
        //fstrcat(filepath, fsp->fsp_name);        
	if (snprintf(filepath, PATH_MAX+1, "/quarantine/%s/%s/%s",
		fsp->conn->client_address, fsp->conn->user,
		fsp->conn->connectpath)
			> PATH_MAX) {return(ENAMETOOLONG);
			}
        fstrcat(filepath, "/"); 
        fstrcat(filepath, fsp->fsp_name);        
	
        /* Don't scan directorys */
        if ( fsp->is_directory )
            return retval;


	if ( !fsp->modified ) {
		#if VSCAN_SOPHOS_VERBOSE_FILE_LOGGING==1
		vscan_syslog("file %s was not modified - not scanned", filepath);
		#endif
		return retval;
	}

	if ( savi_ok ) {
		/* scan only file, do nothing */	
		fstrcpy(client_ip,fsp->conn->client_address);
		rv = vscan_sophos_scanfile(filepath,client_ip);
	}
	/* Delete the file now we don't want it - KS July 02 */
	retval2=remove(filepath);
	if ( ! retval2 ) return retval2;
	return retval;
	#else
	return default_vfs_ops.close(fsp, fd);
	#endif

}

/* VFS initialisation function.  Return initialised vfs_ops structure
   back to SAMBA. */

#if SAMBA_VERSION_MAJOR==3
 /* Samba 3.0 alphaX */
 struct vfs_ops *vfs_init(int *vfs_version, struct vfs_ops *ops)
#else
 /* Samba 2.2.x */
 #if SAMBA_VERSION_MINOR==4   
  /* Samba 2.2.4 */
  struct vfs_ops *vfs_init(int *vfs_version, struct vfs_ops *def_vfs_ops)
 #elif SAMBA_VERSION_MINOR==2
  /* Samba 2.2.2 / Samba 2.2.3 !!! */
  struct vfs_ops *vfs_init(int* Version, struct vfs_ops *ops)
 #elif SAMBA_VERSION_MINOR==1
  /* Samba 2.2.1 */
  struct vfs_ops *vfs_module_init(int *vfs_version)
 #else
  /* Samba 2.2.0 */
  struct vfs_ops *vfs_init(int *vfs_version)
 #endif
#endif
{

	#if SAMBA_VERSION_MINOR==4
	 /* Samba 2.2.4 */
	 struct vfs_ops tmp_ops;
	#endif
 
        openlog("smbd_vscan_sophos", LOG_PID, SYSLOG_FACILITY);

	#if SAMBA_VERSION_MAJOR==3
	 /* Samba 3.0 alphaX */
  	 *vfs_version = SMB_VFS_INTERFACE_VERSION;
	 vscan_syslog("VFS_INIT: vscan_ops - %s loaded\n", module_id);
        #else
	 /* Samba 2.2.x */
	 #if SAMBA_VERSION_MINOR==4
          /* Samba 2.2.4 */
          *vfs_version = SMB_VFS_INTERFACE_VERSION;
          vscan_syslog("VFS_INIT: vscan_ops - %s loaded\n", module_id);
         #elif SAMBA_VERSION_MINOR==2
	  /* Samba 2.2.2 / Samba 2.2.3 !!! */
          *Version = SMB_VFS_INTERFACE_VERSION;
          vscan_syslog("VFS_INIT: vscan_ops - %s loaded\n", module_id);
         #else
          /* Samba 2.2.1 / Samba 2.2.0 */
          *vfs_version = SMB_VFS_INTERFACE_VERSION;
          vscan_syslog("VFS_INIT: &vscan_ops: 0x%8.8x - %s\n",
               &vscan_ops, module_id);
         #endif
	#endif

	#if SAMBA_VERSION_MAJOR==3
         /* Samba 3.0 alphaX */
  	 /* Save a copy of the default ops */

         default_vfs_ops = *ops;

         /* Override our ones */
	 ops->connect = vscan_connect;
	 ops->disconnect = vscan_disconnect;
	 ops->open = vscan_open;
	 ops->close = vscan_close;
	 return(ops); 
        #else
         /* Samba 2.2.x */
	 #if SAMBA_VERSION_MINOR==4
	  /* Samba 2.2.4 */

	  *vfs_version = SMB_VFS_INTERFACE_VERSION;
	  memcpy(&tmp_ops, def_vfs_ops, sizeof(struct vfs_ops));
	  tmp_ops.connect = vscan_connect;
	  tmp_ops.disconnect = vscan_disconnect;
	  tmp_ops.open = vscan_open;
	  tmp_ops.close = vscan_close;
	  memcpy(&vscan_ops, &tmp_ops, sizeof(struct vfs_ops));
	  return(&vscan_ops);

	 #else
          /* Samba 2.2.3-2.2.0 */
          return(&vscan_ops);
	 #endif
	#endif
}

